# Python Imports:
import shutil
import gzip
import sys
import os
import types
import time
import copy
import cPickle
import com.iphrase.content
import iphrase
import iphrase.properties
import iphrase.base.libutil.util
import threading
import mx.DateTime

# Local Imports
import util

# TODO
#   - how to fix docLimit build so that it more quickly aborts the
#     Notes iterator???

# Provides some messages about threads waiting for each other:
VERBOSE = 0

# Not used currently:
DOCS_PER_CACHE_FILE = 500

# How far ahead (in # docs) the LotusNotes thread should get before it
# pauses to give build a chance to catch up:
DOCS_READ_AHEAD = 500

# We set this when iterator is instantiated to record for delta build:
buildStartTime = {}

def getLastBuildTimeFileName(contentSourceName):
  dirName = '%s/buildTimes' % iphrase.properties.get('SITE_HOME')
  if not os.path.exists(dirName):
    os.makedirs(dirName)
  return '%s/%s.txt' % (dirName, contentSourceName)

def castModifyDate(doc):

  """
  Translates the date/time string coming out of Notes API into the
  corresponding Python DateTime object.  This is done in-place to
  doc['__ModifyDate'] if present.
  """

  try:
    s = doc['__ModifyDate']
  except KeyError:
    pass
  else:

    # Seems as if datetime's (at least the Modified header in each
    # Note) come out w/ AM/PM suffix, which mx.DateTime doesn't grok
    # -- so we parse it ourselves:

    try:
      if s[-3:] == ' AM':
        t = mx.DateTime.DateTimeFrom(s[:-3])
      elif s[-3:] == ' PM':
        t = mx.DateTime.DateTimeFrom(s[:-3])
        if t.hour < 12:
          t = t + 12 * mx.DateTime.oneHour
      else:
        t = mx.DateTime.DateTimeFrom(s)
      doc['__ModifyDate'] = t
    except Exception, err:
      if VERBOSE:
        print 'WARNING: Lotus DB: unable to parse modify date "%s": %s; deleting' % \
              (str(s), str(err))
      del doc['__ModifyDate']

    
class LotusDB(com.iphrase.content.Source):

  """
  Implements custom content source on top of exported files from a Lotus
  KB.  Eventually we can be 'more direct' (not require intermediate
  binary file export).
  """

  def __init__(self, fields, randomAccess):

    """
    The fields arg is simple key/value (dict) of the attrs from contentSource.txt.
    """

    com.iphrase.content.Source.__init__(self, fields, randomAccess)
    
    # Where we should read or write to for cache:
    self.dirName = fields.get('cacheDir')
    self.name = fields.get('__name__')

    # debugging info
    self.debug = fields.get('debug')
    self.fieldNames = {}
    self.debugOutput = None
    x = fields.get('debugOutput')
    if x:
      try:
        fh = open(x, 'wb')
        self.debugOutput = fh
        print 'Lotus DB: debugOutput set to %s' % x
      except:
        pass
      

    if self.debug:
      print 'Lotus DB: debug mode enabled'
    self.view = fields.get('view')
    if type(self.view) == types.ListType:
      message = 'ERROR: Lotus DB: view should have only one entry (got: %s)' % str(self.view)
      print message
      raise RuntimeError(message)

    self.formula = fields.get('formula')
    if type(self.formula) == types.ListType:
      message = 'ERROR: Lotus DB: formula should have only one entry (got: %s)' % str(self.formula)
      print message
      raise RuntimeError(message)

    # See if we should load the content live:

    # Mode should be 'cache', 'live':
    self.mode = fields.get('mode')
    if self.mode not in ('live', 'cached'):
      raise RuntimeError('Lotus DB: mode must be live or cached (got "%s")' % str(self.mode))

    if self.mode == 'live-cache-write':
      self.mode = 'live'
      self.cacheWrite = 1
    else:
      self.cacheWrite = 0

    self.serverName = fields.get('server')
    if self.serverName == None:
      self.serverName = ''
    elif type(self.serverName) != types.StringType:
      message = 'ERROR: Lotus DB: server should have only one entry (got: %s)' % str(self.serverName)
      print message
      raise RuntimeError(message)
    elif self.serverName.lower() == 'local':
      self.serverName = ''
    
    self.dbName = fields.get('dbName')
    if type(self.dbName) != types.StringType:
      message = 'ERROR: Lotus DB: dbName should have only one entry (got: %s)' % str(self.dbName)
      print message
      raise RuntimeError(message)
    
    self.username = fields.get('username')
    self.docLimit = fields.get('docLimit')
    if self.docLimit != None:
      self.docLimit = int(self.docLimit)
      print 'Lotus DB: will stop after %d docs...' % self.docLimit
    else:
      self.docLimit = -1

    if self.mode == 'live' and iphrase.base.libutil.util.getCurrentToolName() != 'build':
      print 'Lotus DB: setting mode to "cached" since running tool is not build'
      self.mode = 'cached'

    if 0 and self.mode == 'live' and sys.platform.lower().find('win32') == -1:
      print 'Lotus DB: setting mode to "cached" since OS is not windows'
      self.mode = 'cached'

    self.outputs = []
    l = fields.get('output')
    if l != None:
      if type(l) == types.StringType:
        l = [l]
      for elem in l:
        if elem not in self.outputs:
          self.outputs.append(elem)

    self.required = []
    l = fields.get('required')
    if l != None:
      if type(l) == types.StringType:
        l = [l]
      for elem in l:
        if elem not in self.required:
          self.required.append(elem)
    
    self.currentDoc = None
    self.identifier = fields.get('identifier')
    # Set up remap:
    self.remap = {}
    l = fields.get('rename')

    if l != None:

      if type(l) == types.StringType:
        l = [l]

      for elem in l:
        # Overly simplistic....:
        l2 = elem.strip().split()
        if len(l2) != 2:
          if len(l2) == 3:
            if l2[1].startswith('"') and l2[2].endswith('"'):
              l2 = [l2[0], l2[1][1:] + ' ' + l2[2][:-1]]
            else:
              raise RuntimeError('could not parse rename line "%s"' % elem)
          else:
            raise RuntimeError('could not parse rename line "%s"' % elem)
        self.remap[l2[0]] = l2[1]

    self.static = {}

    # Make all outputs default to None:
    for field in self.outputs:
      self.static[field] = None

    l = fields.get('static')
    if l != None:
      if type(l) == types.StringType:
        l = [l]

      for elem in l:
        l2 = elem.strip().split()
        self.static[l2[0]] = l2[1]

    # print 'Lotus DB: using cached files in directory %s' % self.dirName
    # self.__setFields()

  if 0:
    def __setFields(self):

      fieldsFileName = '%s/fields.pk' % self.dirName

      if not os.path.exists(fieldsFileName):
        print 'Fields file "%s" does not exist ... generating fields now...' % fieldsFileName

        upto = 0

        fields = {}

        while 1:
          fileName = '%s/%s.pk.gz' % (self.dirName, upto)
          upto = upto + 1
          if not os.path.exists(fileName):
            break
          print '  %s...' % os.path.split(fileName)[-1]

          s = gzip.GzipFile(fileName, 'rb').read()

          for d in cPickle.loads(s):
            for field in d.keys():
              fields[field] = 1+ fields.get(field, 0)

        cPickle.dump(fields, open(fieldsFileName, 'wb'), 1)
      else:
        fields = cPickle.load(open(fieldsFileName, 'rb'))
        print 'Loading fields from file "%s"...' % fieldsFileName

      self.fields = fields
      l = fields.items()
      l.sort(lambda x, y: cmp(y[1], x[1]))
      self.outputs = [elem[0] for elem in l]

      for i in xrange(len(self.outputs)):
        key = self.outputs[i]
        if key in self.remap:
          s = ' [remap to %s]' % self.remap[key]
          self.outputs[i] = self.remap[key]
        else:
          s = ''
        print '  %s: %s times%s' % (key, fields.get(key, 'n/a'), s)

  def getIterator(self, delta):
    buildStartTime[self.name] = mx.DateTime.now()
    if self.mode == 'live':
      if self.serverName != '':
        print 'Lotus KB: will pull content live from SERVER %s, DB %s' % \
              (self.serverName, self.dbName)
      else:
        print 'Lotus KB: will pull content live from local DB %s' % \
              self.dbName
      self.iter = LotusDBLiveIterator(self, self.dirName, delta)
    else:
      print 'Lotus KB: will pull content from CACHE %s' % self.dirName
      self.iter = LotusDBCacheIterator(self, self.dirName, delta)
    return self.iter

  def getValue(self, ident, key, rich_doc=None):
    assert ident == self.currentDoc[self.identifier], '%s vs %s' % (ident, self.currentDoc[self.identifier])
    return self.currentDoc.get(key)

  if 0:
    def getValue2(self, ident, key, rich_doc=None):

      if ident != self.identifier:
        if self.iter.isDone():
          return None
        else:
          self.iter.next()
          self.identifier = ident

      return self.currentDoc.get(key)

    if iphrase.base.libutil.util.getCurrentToolName() != 'build':
      getValue = getValue2

  def getOutput(self):
    d = {}
    d.update(self.static)
    for key, value in self.remap.items():
      d[value] = 1

    return d.keys()

class LotusDBLiveIterator(com.iphrase.content.Iterator):

  def __init__(self, parent, dirName, delta):
    self.delta = delta
    self.parent = parent
    self.static = parent.static
    self.view = parent.view
    self.formula = parent.formula
    self.identifier = parent.identifier
    self.debug = parent.debug
    self.nDocs = 0

    for key, value in parent.remap.items():
      if value == self.identifier:
        self.identifierIn = key
        break
    else:
      self.identifierIn = self.identifier

    self.required = parent.required
    if self.identifierIn not in self.required:
      self.required.append(self.identifierIn)
      
    self.dirName = dirName
    self.remap = parent.remap

    self.fileIndex = None
    self.dataIndex = None
    self.fileData = None
    self.__isDone = 0

    # NOTE: we start this "on demand" from isDone; if we start it here
    # we hit a deadlock if you have more than one content source using
    # this connector:
    self.__exportThread = None

  def __runExport(self):

    callback = self.addDoc

    items = {}
    items.update(self.parent.remap)
    for key in self.parent.outputs:
      items[key] = 1

    sinceTime = None
    if self.delta:
      fileName = getLastBuildTimeFileName(self.parent.name)
      if os.path.exists(fileName):
        # Subtract one day just for "margin of safety":
        sinceTime = (mx.DateTime.DateTimeFrom(open(fileName, 'r').read()) - mx.DateTime.oneDay).strftime('%m/%d/%Y %I:%M:%S %p')
        print 'Lotus DB: delta build: retrieving Notes modified after %s' % sinceTime
      else:
        print 'Lotus DB: delta build: could not find last build time [no file "%s"]; indexing all Notes' % fileName

    if 1 or sys.platform.lower().find('win32') != -1:
      import LotusNotes
    else:
      raise RuntimeError('currently connector only runs on WIN32')

    attachmentDirName = '%s/lotusAttachments/' % iphrase.properties.get('SITE_HOME')
    if not os.path.exists(attachmentDirName):
      os.makedirs(attachmentDirName)

    # NOTE: this is not very efficient.  I am enumerating all Notes in
    # the view, and then asking Notes C API to pull back "all Notes"
    # and for each I check (see addDoc below) if that note was
    # actually in the view.  Would be far better to make this view
    # part of the search criteria in the call to NSFSearch, however, I
    # can't figure out how to do that.  There is a tantalizating
    # viewTitle parameter, but, oddly that seems to just be usable by
    # the selection formula and not actually restricts the notes
    # returned:

    # Now expand view --> note ids:
    if self.view is not None:
      self.viewNoteIds = {}
      print
      print 'LotusNotes: getting NoteIDs for view "%s" ...' % self.view
      for noteId in LotusNotes.getViewNoteIds(self.parent.serverName, self.parent.dbName, self.view):
        self.viewNoteIds[noteId] = 1
      print 'LotusNotes: %d notes in view' % len(self.viewNoteIds)
      if '__NoteId' not in items:
        items['__NoteId'] = '__NoteId'
    else:
      self.viewNoteIds = None

    LotusNotes.export(self.parent.serverName,
                               self.parent.dbName,
                               attachmentDirName,
                               callback, items.keys(), sinceTime, self.formula,
                               self.view, self.parent.docLimit)

  def __flushToCache(self):

    # TODO: .new, rename, etc.:
    pass

  def addDoc(self, doc):

    """
    Callback from LotusNotes API to deliver a new doc.  We just append
    to our self.docs list (which build reads from).  And, if we are
    too far ahead of build (too many docs in the queue because we are
    running as our own thread), we wait for build to catch up:
    """

    # If we are indexing only a certain view, then make sure this note
    # is in the view:
    if self.viewNoteIds is not None and \
       doc['__NoteId'] not in self.viewNoteIds:
      return

    # Identifier & any other features marked required must be present
    # in the doc; else, skip it:
    for s in self.required:
      if s not in doc:
        return

    # Un-page any OLE attachments:
    util.unPageAttachments(doc)

    # Start w/ all static fields, including None defaults for all outputs/renames:
    out = {}
    out.update(self.static)

    # Merge in what this doc actually has:
    out.update(doc)

    # TODO: need to be cleaner to do this for all DateTime features,
    # somehow specified in contentSource.txt:
    castModifyDate(out)
    
    for oldName, newName in self.remap.items():
      try:
        value = out[oldName]
      except KeyError:
        pass
      else:
        # don't delete so that a rename can clone a output
        #del out[oldName]
        out[newName] = value

    self.docs.append(out)

    if len(self.docs) > DOCS_READ_AHEAD:
      # Give it a little hysteresis to have more granular thread scheduling for OS:
      while len(self.docs) > DOCS_READ_AHEAD*0.75:
        if VERBOSE:
          print 'Lotus DB: wait for build to catch up [ahead by %d docs]...' % len(self.docs)
        time.sleep(0.25)
    
    return out
    
  def close(self):
    if self.debug:
      print 'Lotus DB: %s FieldNames' % self.parent.name
      kl = self.parent.fieldNames.keys()
      kl.sort()
      for k in kl:
        print '->"%s"  %.2f%%' % (k, self.parent.fieldNames[k]/float(self.nDocs) * 100.0)
    pass

  def isDone(self):

    if self.__exportThread == None:
      self.docs = []
      t = self.__exportThread = threading.Thread(target=self.__runExport, args=())
      t.setDaemon(1)
      t.start()

      while len(self.docs) == 0 and self.__exportThread.isAlive():
        time.sleep(0.25)
    
    if len(self.docs) == 0 and not self.__exportThread.isAlive():
      self.__isDone = 1
    return self.__isDone

  def next(self):

    while 1:

      while len(self.docs) == 0:
        if VERBOSE:
          print 'Lotus DB: build waiting for export thread...'

        if not self.__exportThread.isAlive():
          # We're done -- we can't return "DONE" here -- so return the
          # last doc one more time:
          return ('a', self.parent.currentDoc[self.identifier])
        
        time.sleep(0.25)
        
      # Hopefully it's okay that we do pop and not "fifo"?  Much
      # faster, but is a semi-randomized order as far as build is
      # concerned (which should not be an issue):
      doc = self.parent.currentDoc = self.docs.pop()
      if self.identifier in doc:
        break
      else:
        pass

    if self.parent.debugOutput:
      self.parent.debugOutput.write('%s' % doc.items() + '\n')
    if self.debug:
      self.nDocs += 1
      for k in doc.keys():
        if not k in self.parent.fieldNames:
          self.parent.fieldNames[k] = 0
        self.parent.fieldNames[k] += 1
    return ('a', doc[self.identifier])
  
class LotusDBCacheIterator(com.iphrase.content.Iterator):

  def __init__(self, parent, dirName, delta):

    if delta:
      print 'ERROR: Lotus KB: cached iterator cannot handle delta builds'
      raise RuntimeError('ERROR: Lotus KB: cached iterator cannot handle delta builds')
    
    self.parent = parent
    self.static = parent.static
    self.identifier = parent.identifier
    self.dirName = dirName
    self.fileIndex = None
    self.dataIndex = None
    self.fileData = None
    self.debug = parent.debug
    self.nDocs = 0
    self.__isDone = 0
    if self.parent.docLimit != None:
      self.docLimit = self.parent.docLimit
    else:
      self.docLimit = None

    for key, value in parent.remap.items():
      if value == self.identifier:
        self.identifierIn = key
        break
    else:
      self.identifierIn = self.identifier

    self.required = parent.required
    if self.identifierIn not in self.required:
      self.required.append(self.identifierIn)
      
  def close(self):
    if self.debug:
      print 'Lotus DB: %s FieldNames' % self.parent.name
      kl = self.parent.fieldNames.keys()
      kl.sort()
      for k in kl:
        print '->"%s"  %.2f%%' % (k, self.parent.fieldNames[k]/float(self.nDocs) * 100.0)
    pass

  def isDone(self):
    if self.fileData != None and \
       self.dataIndex == len(self.fileData) and \
       (not os.path.exists('%s/%d.pk.gz' % (self.dirName, 1+self.fileIndex)) or self.docLimit == 0):
      self.__isDone = 1
      
    return self.__isDone

  def next(self):

    while 1:
      if self.fileIndex == None or self.dataIndex >= len(self.fileData):
        if self.fileIndex == None:
          # Start with first file:
          self.fileIndex = 0
        else:
          self.fileIndex = 1 + self.fileIndex

        fileName = '%s/%d.pk.gz' % (self.dirName, self.fileIndex)

        if not os.path.exists(fileName) or self.docLimit == 0:
          self.__isDone = 1
          return None
        else:
          s = gzip.GzipFile(fileName, 'rb').read()
          self.fileData = cPickle.loads(s)
          self.dataIndex = 0

          if self.docLimit != None:
            self.docLimit -= len(self.fileData)
            if self.docLimit < 0:
              self.fileData = self.fileData[:self.docLimit]
              self.docLimit = 0

          # Delete docs that are missing required:
          i = 0
          while i < len(self.fileData):
            doc = self.fileData[i]
            for item in self.required:
              if item not in doc:
                del self.fileData[i]
                break
            else:
              i += 1

      out = self.parent.currentDoc = {}
      out.update(self.static)
      out.update(self.fileData[self.dataIndex])

      castModifyDate(out)
      
      self.dataIndex += 1

      for oldName, newName in self.parent.remap.items():
        try:
          value = out[oldName]
        except KeyError:
          pass
        else:
          # don't delete so that a rename can clone a output
          #del out[oldName]
          out[newName] = value
          
      if self.identifier in out:
        break
      else:
        pass

    if self.parent.debugOutput:
      self.parent.debugOutput.write('%s' % out.items() + '\n')

    if self.debug:
      self.nDocs += 1
      for k in out.keys():
        if not k in self.parent.fieldNames:
          self.parent.fieldNames[k] = 0
        self.parent.fieldNames[k] += 1
    return ('a', out[self.identifier])

def buildPipelineCommit(param):
  
  className = param.getContentSourceField('custom_class')

  if className == 'LotusDB':
    # Only save out last build time if this content source section is
    # "one of ours":
    name = param.getContentSourceField('__name__')
    fileName = getLastBuildTimeFileName(name)
    open(fileName, 'w').write(str(buildStartTime[name]))

def adminStep(param):
  lotusNotesAdminStep(param)

def lotusNotesAdminStep(param):

  # Do not do this if admin was told not to copy the system dir!
  if param.getStep() == 'endBuild' and '-disableCopyBuild' not in sys.argv:

    # Copy back the build timestamp dirs -- needed so ongoing delta
    # builds "do the right thing":
    # TODO: also copy back cache dirs if/when we get that working again:

    for subDirName in ('buildTimes',):

      srcDirName = 'deployment/%s/%s-%s-in_progress/%s' % \
                   (param.lookup('SITE_NAME'),
                    param.lookup('DOMAIN'),
                    param.getTimestamp(),
                    subDirName)

      destDirName = 'deployment/%s/%s/%s' % \
                    (param.lookup('SITE_NAME'),
                     param.lookup('DOMAIN'),
                     subDirName)

      param.addLog('copy back %s --> %s' % (srcDirName, destDirName))
      
      dirCopy(srcDirName, destDirName)


def dirCopy(srcDirName, destDirName):

  """
  Recursively copies directories over, being careful to roll back if
  something fails.
  """

  oldDestDirName = destDirName + '.old'
  newDestDirName = destDirName + '.new'

  # Remove dest.new:
  if os.path.exists(newDestDirName):
    try:
      shutil.rmtree(newDestDirName)
    except Exception, err:
      print 'COPY FAILED: could not copy %s --> %s: unable to remove %s: %s' % \
            (srcDirName, destDirName, newDestDirName, str(err))
      return

  # Copy src --> dest.new:
  try:
    shutil.copytree(srcDirName, newDestDirName)
  except Exception, err:
    print 'COPY FAILED: could not copy %s --> %s: %s' % \
          (srcDirName, destDirName, str(err))
    return

  # Remove dest.old:
  if os.path.exists(oldDestDirName):
    try:
      shutil.rmtree(oldDestDirName)
    except Exception, err:
      print 'COPY FAILED: could not copy %s --> %s: unable to remove %s: %s' % \
            (srcDirName, destDirName, oldDestDirName, str(err))
      return
    
  # Rename dest --> dest.old:
  if os.path.exists(destDirName):
    try:
      os.rename(destDirName, oldDestDirName)
    except Exception, err:
      print 'COPY FAILED: could not copy %s --> %s: unable to rename %s --> %s: %s' % \
            (srcDirName, destDirName, destDirName, oldDestDirName, str(err))
      return

  # Rename dest.new --> dest:
  try:
    os.rename(newDestDirName, destDirName)
  except Exception, err:
    print 'COPY FAILED: could not copy %s --> %s: unable to rename %s --> %s: %s' % \
          (srcDirName, destDirName, newDestDirName, destDirName, str(err))

    # Try to rename dest.old --> dest:
    try:
      os.rename(oldDestDirName, destDirName)
    except Exception, err:
      pass
        
    return

